/******************************************************************************
 * 
 *  textable - TeX tables generator. GNUPlot's data to TeX's tabular.
 *  
 ******************************************************************************
 *
 *  Copyright (C) 2003 Josef Kufner
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 ******************************************************************************
 *
 * TODO
 *	- Format cisel pro ruzne sloupecky
 *	- Nahrazovani v TeX Output "5e3" za "5\\cdot10^{3}" (volitelne)
 *	- Konverze formatu sci -> Eng (5e3 -> 5k)
 *
 *	- uvozovky
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>
#include <ctype.h>

#define	BUF_STEP	128
#define MAX_LEN		512

#define DEBUG_CALC	0	/* 1 or 0 */

#define PRINT_ERR_PRG(str, arg...)	{fprintf(stderr, "%s: " str "\n", prg_name, ##arg);}
#define PRINT_ERR_FILE(str, arg...)	{fprintf(stderr, "%s:%4d: " str "\n", in_file, line_no, ##arg);}

#define PRINT_DEBUG(arg...)		{if (debug) PRINT_ERR_FILE(##arg); }
#define PRINT_VERBOSE(arg...)		{if (verbose) PRINT_ERR_FILE(##arg); }
#define PRINT_VERBOSE_PRG(arg...)	{if (verbose) PRINT_ERR_PRG(##arg); }

#define PRINT_DEBUG_POINT()		{PRINT_ERR_PRG("Reached line %d.", __LINE__)}

typedef struct struct_var_format {
	enum {
		REAL = 0, SCI, ENG, MAX_MODE	/* MAX_MODE isn't real mode ! */
	} mode;

	int digits;
	
	struct struct_var_format *next;
} VAR_FORMAT;

const char *mode_string[4] = { "real", "sci", "sci", NULL };
VAR_FORMAT var_format_default = { REAL, 2, NULL };

typedef struct struct_var_list {
	char name[MAX_LEN];
	char val[MAX_LEN];
	char expr[MAX_LEN];
/*	VAR_FORMAT *format;*/
	struct struct_var_list *next;
} VAR_LIST;

VAR_LIST *var_list = NULL;
VAR_LIST *expr_list = NULL;
VAR_LIST *add_expr_list = NULL;
VAR_FORMAT *var_format_list = NULL;

int verbose = 0, debug = 0;
char *in_file = NULL, *out_file = NULL, *out_plain_file;
char *prg_name = NULL;

FILE *f_in = NULL, *f_out = NULL, *f_plain_out = NULL;
int line_no = 0;
int exit_val = 0;


/******************************************************************************
 *	malloc_c -- enhanced malloc (+ error check)
 *****************************************************************************/
void *malloc_c (const size_t size) {
	void *mem;
	
	mem = malloc(size);
	if (!mem) {
		PRINT_ERR_PRG("Error while alocating memory: %s",
				strerror(errno));
		exit(2);
	}
	return mem;
}

/******************************************************************************
 *	read_line -- reads one line from input
 *****************************************************************************/
int read_line (FILE *f, char **line) {
	char *buf = NULL, ch;
	char *line_buf = NULL;
	int i;
	
	line_no++;
	
	/* Free old string */
	if (line_buf) {
		free(line_buf);
		line_buf = NULL;
	}
	
	/* initial buf size */
	line_buf = (char *) malloc_c(BUF_STEP * sizeof(char));
	line_buf[0] = '\0';
	*line = line_buf;
	
	/* reading */
	for (i = 0; !feof(f) && (ch = getc(f)) != '\n'; i++) {

		/* replace \t with space - only for debug */
		ch = ch == '\t' ? ' ' : ch;
		
		/* enlarge buffer */
		if (i % BUF_STEP == BUF_STEP - 1) {
			PRINT_VERBOSE_PRG("Line %d too long -- enlarging input buffer to %d bytes.",
						line_no, (i + BUF_STEP + 1) * sizeof(char));
			/* backup */
			buf = (char *) malloc_c(i * sizeof(char) + 1);
			memcpy(buf, line_buf, i * sizeof(char));
			free(line_buf);
			
			/* copy back */
			line_buf = (char *) malloc_c((i + BUF_STEP) * sizeof(char) + 1);
			memcpy(line_buf, buf, i * sizeof(char));
			*line = line_buf;
			free(buf);
		}

		/* add char to buffer */
		line_buf[i] = ch;
		line_buf[i+1] = '\0';
	}
	
	/* return value */
	if (feof(f))
		return 0;
	return 1;
	
}

/******************************************************************************
 *	on_Exit -- call this when exiting
 *****************************************************************************/
void on_Exit (void) {

	PRINT_DEBUG_POINT();

	/* zavirani souboru */
	if (f_in)
		fclose(f_in);
	if (f_out)
		fclose(f_out);
	if (f_plain_out)
		fclose(f_plain_out);

	PRINT_DEBUG_POINT();

	/* uvolneni pameti */
	if (in_file)
		free(in_file);
	if (out_file)
		free(out_file);
	if (out_plain_file)
		free(out_plain_file);

	/* TODO free all lists */
	
	PRINT_DEBUG_POINT();

	/* Say Bye */
	PRINT_VERBOSE_PRG("%s", "Bye.");
	
	if (prg_name)
		free(prg_name);

	PRINT_DEBUG_POINT();

	exit(exit_val);
}

/******************************************************************************
 *	calc -- vypocet
 *****************************************************************************/
int calc (VAR_LIST *variable) {
	char *expr;
	char *new_expr;
	
	FILE *f_calc;
	char *cmd, result[MAX_LEN] = "(unknown)";
	int ret;
/*	VAR_FORMAT *format = NULL;
*/		
	
	new_expr = (char *) malloc_c(strlen(variable->expr) + 1);
	strcpy(new_expr, variable->expr);
	
	if (strlen(variable->val) == 0) {
		int done = 0;
		
		/* vezmem vyraz a postupne resime */
		
		/* cyklicky dosazujem za promenne vyrazy */
		while (!done) {
			char var[MAX_LEN];
			char *subst;
			int i, undef_var;
			
			expr = new_expr;
#if DEBUG_CALC
			fprintf(stderr, "> %s\n", expr);
			sleep(1);
#endif
			
			/* nejdrive najdem promennou */
			if (*expr != '\0' && (expr = strchr(expr, '$')) != NULL) {
				char *tmp_expr = NULL;
				int var_pos = 0;
				VAR_LIST *curr;
				
				/* nasli -- dosadime za ni dalsi vyraz */
				expr++;
				var_pos = expr - new_expr - 1;
				
				/* zjisteni jmena promenne */
				i = 0;
				while (*expr != '\0' && (isalnum(*expr) || *expr == '_')) {
					if (i < MAX_LEN - 1) {
						var[i++] = *expr++;
						var[i] = '\0';
					}
				}

#if DEBUG_CALC
				fprintf(stderr, "var: $%s (var_pos=%d)\n", var, var_pos);
				sleep(1);
#endif
				
				/* nalezeni nahrazujiciho vyrazu */
				undef_var = 1;
				for (curr = (isdigit(var[0]) ? var_list : expr_list)->next;
						curr != NULL;
						curr = curr->next)
				{
					if (strcmp(var, curr->name) == 0) {
						subst = curr->expr;
						undef_var = 0;
						break;
					}
				}
				
				/* Pokud neni nalezeno*/
				if (undef_var) {
					fprintf(stderr, "%s:%d: Fatal: Undefined variable \"$%s\"\n",
							in_file, line_no, var);
					exit(1);
				}

#if DEBUG_CALC
				fprintf(stderr, "S: %s\n", subst);
				sleep(1);
#endif
				
				/* alokace pameti pro nahrazeni */
				tmp_expr = (char *) malloc_c(strlen(new_expr) - (strlen(var) + 1) + strlen(subst) + 1);
				
				/* prekopirovani s nahrazenim */
				strncpy(tmp_expr, new_expr, var_pos);
				tmp_expr[var_pos] = '\0';
				strcat(tmp_expr, subst);
				strcat(tmp_expr, new_expr + var_pos + strlen(var) + 1);
				
#if DEBUG_CALC
				fprintf(stderr, "tmp: %s, varlen=%d\n", tmp_expr, strlen(var));
				sleep(1);
#endif
				
				/* uvolneni pameti po puvodnim vyrazu */
				free(new_expr);

				/* aktualizace pointeru */
				new_expr = tmp_expr;

				/* a pujdem znovu */
#if DEBUG_CALC
				fprintf(stderr, "new: %s\n", new_expr);
				sleep(1);
#endif
				done = 0;
			} else {
				/* nenasli -- hotovo */
				done = 1;
			}
		}

		/* Kontrola zda je prirazen format */
/*		format = variable->format ? variable->format : &var_format_default;
*/

		/* uz nezbyly zadne promenne, tak to spocitame */
/*		sprintf(cmd, "calc -p '"
				"config(\"mode\", \"%s\"); "
				"config(\"display\", %d); "
				"config(\"tilde\", 0); "
				"config(\"leadzero\", 1); "
				"config(\"fullzero\", 1); "
				"%s'",
				mode_string[format->mode >= 0 && format->mode < MAX_MODE ?
					format->mode : REAL],
				format->digits >= 0 ? format->digits : 0,
				new_expr);
*/
		sprintf(cmd, "calc -p '"
				"%s'",
				new_expr);

		PRINT_DEBUG("Calculating: \"%s\"...", new_expr);
		PRINT_DEBUG("popen: %s", cmd);
		
		if((f_calc = popen(cmd, "r")) == NULL) {
			PRINT_ERR_PRG("Fatal: Can't run \"%s\": %s", cmd, strerror(errno));
			exit(-127);
		}
	
		while (!feof(f_calc)) {
			fgets(result, MAX_LEN, f_calc);
			/*PRINT_DEBUG("calc: %s", result);*/
		}
		
		ret = pclose(f_calc);
		/*PRINT_DEBUG("calc exit code: %d", ret);*/

		/* odstranime <CR> */
		{
			char *cr;
			if ((cr = strchr(result, '\n')) != NULL)
				*cr = '\0';
		}

		strcpy(variable->val, result);
		PRINT_DEBUG("     Result:   = \"%s\"", variable->val);

	} else {
		/* je jiz spocitano */
		return 0;
	}

	/* Hotovo */
	free(new_expr);
	return 0;	
}

/******************************************************************************
 *	main
 *****************************************************************************/
int main (int argc, char **argv) {
	int opt;
	char *line = NULL;
	int hd_no = 0;
	int add_no = 0;

	/* global lists */
	VAR_LIST *expr_list_curr = NULL;
	VAR_LIST *add_expr_list_curr = NULL;
			
	/* pomocne promenne */	
	int col_no, i;
	char *data = NULL, str[MAX_LEN];
	char *cmd = NULL, cmd_fn[32];

	/* pointery na aktualni polozky v seznamu */
	VAR_LIST *var_list_curr = NULL;
	VAR_LIST *var_list_next = NULL;
	VAR_LIST *add_curr = NULL;
				

	
/*	atexit(on_Exit);*/

	prg_name = (char *) malloc_c((strlen(argv[0]) + 1) * sizeof(char));
	strcpy(prg_name, argv[0]);
	
	/* Args */
	while ((opt = getopt(argc, argv, "i:o:p:vd")) != -1) {
		switch(opt) {
			case 'i':
				in_file = (char *) malloc_c((strlen(optarg) + 1) * sizeof(char));
				strcpy(in_file, optarg);
				break;
			case 'o':
				out_file = (char *) malloc_c((strlen(optarg) + 1) * sizeof(char));
				strcpy(out_file, optarg);
				break;
			case 'p':
				out_plain_file = (char *) malloc_c((strlen(optarg) + 1) * sizeof(char));
				strcpy(out_plain_file, optarg);
				break;
			case 'v':
				verbose = 1;
				break;
			case 'd':
				debug = 1;
				break;
			case 'h':
			case '?':
				fprintf(stderr, "usage: %s [-i <input_file>] [-o <output_LaTeX_file>] "
						"[-p <output_plain_text_file>] [-v]\n",
						prg_name);
				return -128;
				break;
		}
	}
	
	/* Verbose info */
	PRINT_VERBOSE_PRG("input:  \"%s\"", in_file ? in_file : "<stdin>");
	PRINT_VERBOSE_PRG("output: \"%s\"", out_file ? out_file : "<stdout>");

	/* Input */
	if (in_file) {
		f_in = fopen(in_file, "rt");
		if (!f_in) {
			PRINT_ERR_PRG("Error while opening input file: %s", strerror(errno));
			return 1;
		}
	} else {
		f_in = stdin;
	}

	/* Output */
	if (out_file) {
		f_out = fopen(out_file, "wt");
		if (!f_out) {
			PRINT_ERR_PRG("Error while opening LaTeX output file: %s", strerror(errno));
			return 1;
		}
	} else {
		f_out = stdout;
	}

	/* Plain Output */
	if (out_plain_file) {
		f_plain_out = fopen(out_plain_file, "wt");
		if (!f_plain_out) {
			PRINT_ERR_PRG("Error while opening plain text output file: %s\n", strerror(errno));
			return 1;
		}
	}

	/* Pripraveni seznamu AddExpr */
	add_expr_list = (VAR_LIST *) malloc_c(sizeof(VAR_LIST));
	add_expr_list_curr = add_expr_list;
	add_expr_list->next = NULL;		/* prvni polozku nepouzijem */
	
	/* Pripraveni seznamu Expr */
	expr_list = (VAR_LIST *) malloc_c(sizeof(VAR_LIST));
	expr_list_curr = expr_list;
	expr_list->next = add_expr_list->next;	/* prvni polozku nepouzijem */
		
			
	/* Generate */
	/* Head */
	fprintf(f_out, "%% This file was generated from \"%s\" by TeX Table\n",
			(in_file != NULL) ? in_file : "<stdin>");
	
	while (read_line(f_in, &line)) {

		if (line[0] == '#') {
			if (line[1] == '>') {
				/* Prikazy */
				
				/* posun na zacatek prikazu */
				cmd = line + 2;
				
				/* Command */
				cmd_fn[0] = '\0';
				for (i = 0; i < 32 - 1 && isalpha(*cmd); i++) {
					cmd_fn[i] = *cmd;
					cmd_fn[i + 1] = '\0';
					cmd++;
				}

				/* Skip whitespace */
				while (isspace(*cmd)) {
					cmd++;
				}

			
				/* Provedeni prikazu */
				/* FORMAT */
				if (strcmp(cmd_fn, "FORMAT") == 0) {
					PRINT_DEBUG("Table format: \"%s\"", cmd);
					fprintf(f_out, "\\begin{center}\n\\begin{tabular}{%s}\n\\hline\n", cmd);
				}	
				
				/* HD */
				else if (strcmp(cmd_fn, "HD") == 0) {
					PRINT_DEBUG("Header %d: \"%s\"", hd_no, cmd);
					fprintf(f_out, "%s%s\n", hd_no > 0 ? "\t& " : "\t  ", cmd);
					hd_no++;
				}
				
				/* HEND */
				else if (strcmp(cmd_fn, "HEND") == 0) {
					PRINT_DEBUG("End of header (cols %d).", hd_no);
					fprintf(f_out, "\t\\\\\n\\hline\n");
					hd_no = 0;
				}

				/* HLINE */
				else if (strcmp(cmd_fn, "HLINE") == 0) {
					PRINT_DEBUG("%s", "HLine");
					fprintf(f_out, "\\hline\n");
				}
			
				/* ADD */
				else if (strcmp(cmd_fn, "ADD") == 0) {
					char var[MAX_LEN];
					
					/* Vytvoreni polozky seznamu */
					add_expr_list_curr->next = (VAR_LIST *) malloc_c(sizeof(VAR_LIST));
					add_expr_list_curr = add_expr_list_curr->next;
					add_expr_list_curr->next = NULL;

					/* Naplneni hodnotama */
					sprintf(var, "_%d", ++add_no);
					strcpy(add_expr_list_curr->name, var);
					strcpy(add_expr_list_curr->expr, cmd);

					/* vypis */
					PRINT_VERBOSE("ADD: %s = \"%s\"",
								add_expr_list_curr->name,
								add_expr_list_curr->expr);
				}
				
				/* CALC */
				else if (strcmp(cmd_fn, "CALC") == 0) {
					char var[MAX_LEN];
					int i;
					
					/* Vytvoreni polozky seznamu */
					expr_list_curr->next = (VAR_LIST *) malloc_c(sizeof(VAR_LIST));
					expr_list_curr = expr_list_curr->next;
					expr_list_curr->next = add_expr_list->next;

					/* prvni parametr -- jmeno promenne */
					i = 0;
					while( i < MAX_LEN - 1 && (isalnum(*cmd) || *cmd == '_')) {
						var[i++] = *(cmd++);
						var[i] = '\0';
					}

					/* Skip whitespace */
					while (isspace(*cmd) || *cmd == '=') {
						cmd++;
					}
					
					/* Naplneni hodnotama */
					strcpy(expr_list_curr->name, var);
					strcpy(expr_list_curr->expr, cmd);

					/* vypis */
					PRINT_VERBOSE("CALC: %s = \"%s\"",
								expr_list_curr->name,
								expr_list_curr->expr);
				}
				
				/* Neznamy */
				else {
					PRINT_ERR_FILE("Warning: Unknown command \"%s\" -- ignoring", cmd_fn);
					exit_val++;
				}
			} else {
				/* Komentar */
				PRINT_DEBUG("%s", "Skipping comment.");
			}
		} else {
			/* Data */
			data = line;

			/* Pripraveni seznamu Var */
			var_list = (VAR_LIST *) malloc_c(sizeof(VAR_LIST));
			var_list_curr = var_list;
			var_list->next = NULL;		/* prvni polozku nepouzijem */
	
			/* dekodovani radky */
			col_no = 0;
			while (*data != '\0') {
				int i = 0;
		
				/* na radce jsou jeste hodnoty */
				/* skip whitespace */
				while (isspace(*data)) {
					data++;
				}

				/* Kontrola konce radky */
				if (*data == '\0')
					continue;

				/* nacteni retezce */
				/* TODO uvozovky */
				i = 0;
				while (*data != '\0' && !isspace(*data)) {
					str[i++] = *data++;
					str[i] = '\0';
				}
				col_no++;


				/* vytvoreni polozky seznamu */
				var_list_curr->next = (VAR_LIST *) malloc_c(sizeof(VAR_LIST));
				var_list_curr = var_list_curr->next;
				var_list_curr->next = NULL;
				
				/* nastaveni hodnot polozky */
				strcpy(var_list_curr->expr, str);
				strcpy(var_list_curr->val, "");
				sprintf(var_list_curr->name, "%d", col_no);
/*				var_list_curr->format = NULL;	*/
				PRINT_DEBUG("Reading data: col %d", col_no);
			}

			/* kopii add_expr_list na konec -- pouze pokud byl nejaky sloupec zadan */
			for (add_curr = add_expr_list->next; col_no > 0 && add_curr != NULL; add_curr = add_curr->next) {
				
				col_no++;

				/* vytvoreni polozky seznamu */
				var_list_curr->next = (VAR_LIST *) malloc_c(sizeof(VAR_LIST));
				var_list_curr = var_list_curr->next;
				var_list_curr->next = NULL;
				
				/* nastaveni hodnot polozky */
				strcpy(var_list_curr->expr, add_curr->expr);
				strcpy(var_list_curr->val, "");
				sprintf(var_list_curr->name, "%d", col_no);
				PRINT_DEBUG("Generating additional data: col %d", col_no);
			}
			
			/* Vypocty na radce */
			for (var_list_curr = var_list->next;
					var_list_curr != NULL;
					var_list_curr = var_list_curr->next)
			{
				/* Je hodnota jiz vypoctena? */
				if (strlen(var_list_curr->val) == 0) {
					/* Neni -> spocitame */
					calc(var_list_curr);
				}
			}

			/* zapis radky */
			if (col_no > 0) {
				for (var_list_curr = var_list->next;
						var_list_curr != NULL;
						var_list_curr = var_list_curr->next)
				{
					/* vystup do plain text */
					if (f_plain_out) {
						fprintf(f_plain_out, "%s%s",
							var_list_curr != var_list->next ? "\t" : "",
							var_list_curr->val);
					}
					
					/* nahrazeni '.' -> ',' */
					{
						char *dot;
						while ((dot = strchr(var_list_curr->val, '.')) != NULL)
							*dot = ',';
					}
					
					/* vystup do LaTeX */
					fprintf(f_out, "\t%s%s",
							var_list_curr != var_list->next ? "& " : "",
							var_list_curr->val);

				}

				fprintf(f_out, "\t\\\\\n");
				if (f_plain_out) {
					fprintf(f_plain_out, "\n");
				}
			}
		
			/* Likvidace seznamu */
			for (var_list_curr = var_list->next;
					var_list_curr != NULL;
					var_list_curr = var_list_next)
			{
				var_list_next = var_list_curr->next;
				free(var_list_curr);
			}
		}
	}

	/* Foot */
	fprintf(f_out, "\\hline\n\\end{tabular}\n\\end{center}\n");
	fflush(f_out);
	PRINT_DEBUG_POINT();
	return;
}


